import abc
from typing import Iterable

from wai.logging import LOGGING_WARNING

from seppl import InputConsumer, PluginWithLogging
from seppl import SessionHandler, Session


class Writer(PluginWithLogging, InputConsumer, SessionHandler, abc.ABC):
    """
    Ancestor of classes that write data.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the handler.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self._session = None
        self._last_input = None

    @property
    def session(self) -> Session:
        """
        Returns the current session object

        :return: the session object
        :rtype: Session
        """
        return self._session

    @session.setter
    def session(self, s: Session):
        """
        Sets the session object to use.

        :param s: the session object
        :type s: Session
        """
        self._session = s

    def _has_input_changed(self, current_input: str = None, update: bool = False) -> bool:
        """
        Checks whether the current input is different from the last one we processed.

        :param current_input: the current input, uses the current_input from the session if None
        :type current_input: str
        :param update: whether to update the last input immediately
        :type update: bool
        :return: True if input has changed
        :rtype: bool
        """
        if current_input is None:
            current_input = self.session.current_input
        result = self._last_input != current_input
        if update:
            self._update_last_input(current_input)
        return result

    def _update_last_input(self, current_input: str):
        """
        Updates the last input that was processed.

        :param current_input: the "new" last input
        :type current_input: str
        """
        self._last_input = current_input

    def _output_needs_changing(self, current_output: str, target: str, ext: str) -> bool:
        """
        Checks whether the output needs changing.

        :param current_output: the current output
        :type current_output: str
        :param target: the output target
        :type target: str
        :param ext: the extension for the output file, incl dot
        :type ext: str
        :return: True if the output needs to change
        :rtype: bool
        """
        raise NotImplementedError()


class StreamWriter(Writer, abc.ABC):
    """
    Ancestor for classes that write data one record at a time.
    """

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        raise NotImplementedError()


class BatchWriter(Writer, abc.ABC):
    """
    Ancestor of classes that write data all at once.
    """

    def write_batch(self, data: Iterable):
        """
        Saves the data in one go.

        :param data: the data to write
        :type data: Iterable
        """
        raise NotImplementedError()



from ._plugin import Plugin, PluginWithLogging, OutputProducer, InputConsumer, check_compatibility
from ._plugin import LoggingHandler, Initializable, init_initializable
from ._plugin import AliasSupporter, get_aliases, has_aliases, get_all_names
from ._args import split_cmdline, split_args, args_to_objects, is_help_requested, enumerate_plugins, escape_args, unescape_args, resolve_handler
from ._registry import Registry, MODE_DYNAMIC, MODE_EXPLICIT, MODES
from ._class_registry import ClassListerRegistry, get_class_lister, DEFAULT
from ._entry_points import generate_entry_points
from ._help import generate_plugin_usage, generate_help, HELP_FORMATS, HELP_FORMAT_TEXT, HELP_FORMAT_MARKDOWN
from ._types import get_class, get_class_name, classes_to_str, AnyData
from ._metadata import MetaDataHandler, add_metadata, get_metadata
from ._session import Session, SessionHandler
from ._placeholders import PlaceholderSupporter, InputBasedPlaceholderSupporter
from ._placeholders import expand_placeholders, placeholder_list, placeholder_help, add_placeholder, load_user_defined_placeholders



import importlib
import inspect
import os
import sys
import traceback

from typing import List, Union, Optional, Dict
from pkg_resources import working_set

from ._plugin import get_all_names, get_aliases, Plugin
from ._types import get_class, get_class_name


MODE_EXPLICIT = "explicit"
MODE_DYNAMIC = "dynamic"
MODES = [
    MODE_EXPLICIT,
    MODE_DYNAMIC,
]

DEFAULT = "DEFAULT"
""" the placeholder for the default modules in the environment variable. """


class Registry:
    """
    Registry for managing plugins derived from seppl.Plugin class.

    Entry points must have the format:

    1. explicit mode

    Each plugin has to be listed by (unique) name ("plugin_name"), with the
    module it resides in ("plugin_module") and its class name ("plugin_class"):

    entry_points={
        "group": [
            "plugin_name=plugin_module:plugin_class",
        ]
    }

    2. dynamic mode

    Only the superclass is listed

    entry_points={
        "group": [
            "unique_string=plugin_module:superclass_name",
        ]
    }

    If a super class ("superclass_name1") has classes in multiple modules
    ("plugin_module1", "plugin_module2"), then these modules need to be
    listed separately:

    entry_points={
        "group": [
            "unique_string1=plugin_module1:superclass_name1",
            "unique_string2=plugin_module2:superclass_name1",
            "unique_string3=plugin_module3:superclass_name2",
            ...
        ]
    }

    When enforcing uniqueness, the "plugin_name" must be unique across all plugins.

    You can use the placeholder DEFAULT in the environment variable to represent the
    default modules. This avoids having to update the env variable whenever the default
    modules change.

    With `excluded_modules` and `excluded_env_modules` you can avoid modules being
    included in the registry, e.g., when generating help output for dependent modules
    (and you don't want all the base plugins to be output as well).
    """

    def __init__(self, mode: Optional[str] = MODE_EXPLICIT,
                 default_modules: Optional[Union[str, List[str]]] = None,
                 env_modules: Optional[str] = None,
                 excluded_modules: Optional[Union[str, List[str]]] = None,
                 excluded_env_modules: Optional[str] = None,
                 enforce_uniqueness: bool = True):
        """
        Initializes the registry. default_modules and env_modules are used as fallback option
        in case no plugins are being obtained from entry_points.

        :param mode: how the registry parses the entry_points
        :type mode: str
        :param default_modules: the default modules to use for registering plugins, comma-separated string of module names or list of module names
        :type default_modules: str or list
        :param env_modules: the environment variable to retrieve the modules from (overrides default ones)
        :type env_modules: str
        :param excluded_modules: the modules to exclude from registering plugins, comma-separated string of module names or list of module names, ignored if None
        :type excluded_modules: str or list
        :param excluded_env_modules: the environment variable to retrieve the excluded modules from (overrides manually set ones)
        :type excluded_env_modules: str
        :param enforce_uniqueness: whether plugin names must be unique
        :type enforce_uniqueness: bool
        """
        if mode not in MODES:
            raise Exception("Unknown mode: %s" % mode)

        self._plugins = dict()
        self._all_plugins = dict()
        self._all_aliases = set()
        self._custom_modules = None
        self._default_modules = None
        self._env_modules = None
        self._excluded_modules = None
        self._excluded_env_modules = None

        self.mode = mode
        self.default_modules = default_modules
        self.env_modules = env_modules
        self.excluded_modules = excluded_modules
        self._excluded_env_modules = excluded_env_modules
        self.enforce_uniqueness = enforce_uniqueness

    def has_env_modules(self) -> bool:
        """
        Checks whether an environment variable for modules is set.

        :return: True if set
        :rtype: bool
        """
        return (self._env_modules is not None) \
               and (len(self._env_modules) > 0) \
               and (os.getenv(self._env_modules) is not None) \
               and (len(os.getenv(self._env_modules)) > 0)

    def has_excluded_env_modules(self) -> bool:
        """
        Checks whether an environment variable for excluded modules is set.

        :return: True if set
        :rtype: bool
        """
        return (self._excluded_env_modules is not None) \
               and (len(self._excluded_env_modules) > 0) \
               and (os.getenv(self._excluded_env_modules) is not None) \
               and (len(os.getenv(self._excluded_env_modules)) > 0)

    @property
    def default_modules(self) -> Optional[List[str]]:
        """
        Returns the default modules.

        :return: the modules
        :rtype: list
        """
        return self._default_modules

    @default_modules.setter
    def default_modules(self, modules: Optional[Union[str, List[str]]]):
        """
        Sets/unsets the default modules to use. Clears the plugin cache.

        :param modules: the list of modules to use, None to unset
        :type modules: list
        """
        if modules is None:
            modules = ""
        if isinstance(modules, str):
            modules = [x.strip() for x in modules.split(",")]
        elif isinstance(modules, list):
            modules = modules[:]
        else:
            raise Exception("default_modules must be either str or list, but got: %s" % str(type(modules)))
        if len(modules) == 0:
            raise Exception("No default modules defined!")
        self._default_modules = modules
        self._plugins = dict()

    @property
    def excluded_modules(self) -> Optional[List[str]]:
        """
        Returns the excluded modules.

        :return: the modules
        :rtype: list
        """
        return self._excluded_modules

    @excluded_modules.setter
    def excluded_modules(self, modules: Optional[Union[str, List[str]]]):
        """
        Sets/unsets the excluded modules to use. Clears the plugin cache.

        :param modules: the list of modules to exclude, None to unset
        :type modules: list
        """
        if modules is None:
            modules = ""
        if isinstance(modules, str):
            modules = [x.strip() for x in modules.split(",")]
        elif isinstance(modules, list):
            modules = modules[:]
        else:
            raise Exception("excluded_modules must be either str or list, but got: %s" % str(type(modules)))
        self._excluded_modules = modules
        self._plugins = dict()

    @property
    def env_modules(self) -> Optional[str]:
        """
        Returns the environment modules (if any).

        :return: the modules, None if none set
        :rtype: str
        """
        return self._env_modules

    @env_modules.setter
    def env_modules(self, modules: Optional[str]):
        """
        Sets/unsets the environment variable with the modules to use. Clears the plugin cache.

        :param modules: the environment variable with the modules to use, None to unset
        :type modules: str
        """
        self._env_modules = modules
        self._plugins = dict()

    @property
    def custom_modules(self) -> Optional[List[str]]:
        """
        Returns the custom modules (if any).

        :return: the modules, None if none set
        :rtype: list
        """
        return self._custom_modules

    @custom_modules.setter
    def custom_modules(self, modules: Optional[Union[str, List[str]]]):
        """
        Sets/unsets the custom modules to use. Clears the plugin cache.

        :param modules: the list of modules to use, None to unset
        :type modules: list
        """
        self._custom_modules = modules
        self._plugins = dict()

    def _expand_default_modules_placeholder(self, m: str) -> str:
        """
        Expands the DEFAULT modules placeholder in the comma-separated modules string.

        :param m: the modules string to expand
        :type m: str
        :return: the expanded modules string
        :rtype: str
        """
        if DEFAULT in m:
            if len(self.default_modules) > 0:
                m = m.replace(DEFAULT, ",".join(self.default_modules))
            else:
                m = m.replace(DEFAULT, "")
        return m

    def actual_fallback_modules(self) -> List[str]:
        """
        Returns list the of modules to fall back on.
        Precedence: custom_modules > env_modules > default_modules

        :return: the list of modules
        :rtype: list
        """
        if (self._custom_modules is not None) and (len(self._custom_modules) > 0):
            return self._custom_modules

        if self.has_env_modules():
            m = self._expand_default_modules_placeholder(os.getenv(self.env_modules))
            return [x.strip() for x in m.split(",")]

        return self.default_modules[:]

    def actual_excluded_modules(self) -> List[str]:
        """
        Returns list the of excluded modules.
        Precedence: excluded_env_modules > excluded_modules

        :return: the list of modules
        :rtype: list
        """
        if self.has_excluded_env_modules():
            m = self._expand_default_modules_placeholder(os.getenv(self._excluded_env_modules))
            return [x.strip() for x in m.split(",")]

        return self.excluded_modules[:]

    def is_excluded(self, o: Union[str, Plugin]) -> bool:
        """
        Checks whether the plugin object or plugin classname fall into an excluded module.

        :param o: the plugin classname/object to check
        :return: True if excluded, otherwise False
        """
        result = False
        excl = self.actual_excluded_modules()
        if isinstance(o, str):
            classname = o
        else:
            classname = get_class_name(o)
        for m in excl:
            if classname.startswith(m + "."):
                result = True
                break
        return result

    def is_alias(self, name: str) -> bool:
        """
        Checks whether the plugin name is an alias.

        :param name: the plugin name to check
        :type name: str
        :return: True if an alias
        :rtype: bool
        """
        return name in self._all_aliases

    @property
    def all_aliases(self) -> List[str]:
        """
        Returns a sorted list of all known aliases.

        :return: the list of all known aliases
        :rtype: list
        """
        return sorted(list(self._all_aliases))

    def _register_plugin(self, d: Dict[str, Plugin], o: Plugin) -> bool:
        """
        Adds the plugin to the registry dictionary under its name.
        Ensures that names are unique and throws an Exception if not.

        :param d: the dictionary to add the handler to
        :type d: dict
        :param o: the plugin to register
        :type o: CommandlineHandler
        :return: whether it was registered
        :rtype: bool
        """
        if self.is_excluded(o):
            return False
        names = get_all_names(o)
        present = any([x in self._all_plugins for x in names])
        if self.enforce_uniqueness and present:
            for name in names:
                if get_class_name(self._all_plugins[name]) != get_class_name(o):
                    raise Exception("Duplicate plugin name encountered: name=%s, existing type=%s, new type=%s)"
                                    % (name, str(type(self._all_plugins[name])), str(type(o))))
        else:
            for name in names:
                self._all_plugins[name] = o
                d[name] = o

        # record aliases
        self._all_aliases.update(get_aliases(o))

        return True

    def _init_plugin_class(self, c):
        """
        Initializes the class to restrict the plugins to.

        :param c: the class, uses Plugin if None
        :return: the plugin class
        """
        if c is None:
            c = Plugin
        elif not issubclass(c, Plugin):
            raise Exception("Class '%s' is not derived from '%s'!" % (str(c), str(Plugin)))
        return c

    def _register_from_module(self, m: str, c: Optional = None):
        """
        Locates all the classes implementing the specified class in the module and
        adds them to the dictionary.

        :param m: the module to look for classes
        :type m: str
        :param c: the class that the plugins must be, any class derived from Plugin if None
        """
        c = self._init_plugin_class(c)
        result = dict()

        module = importlib.import_module(m)
        for att_name in dir(module):
            if att_name.startswith("_"):
                continue
            att = getattr(module, att_name)
            if inspect.isclass(att) and issubclass(att, c):
                try:
                    p = att()
                    if self._register_plugin(result, p):
                        names = get_all_names(p)
                        for name in names:
                            result[name] = p
                        # record aliases
                        self._all_aliases.update(get_aliases(p))
                except NotImplementedError:
                    pass
                except:
                    print("Problem encountered instantiating: " + m + "." + att_name, file=sys.stderr)
                    traceback.print_exc()

        return result

    def _register_from_modules(self, c: Optional = None):
        """
        Locates all the classes implementing the specified class and adds them to the dictionary.

        :param c: the class that the plugins must be, any class derived from Plugin if None
        """
        c = self._init_plugin_class(c)
        result = dict()

        for m in self.actual_fallback_modules():
            result.update(self._register_from_module(m, c))

        return result

    def _register_from_entry_point(self, group: str, c: Optional = None) -> Dict[str, Plugin]:
        """
        Generates a dictionary (name/object) for the specified entry_point group.

        :param group: the entry_point group to generate dictionary for
        :type group: str
        :return: the generated dictionary
        :rtype: dict
        """
        c = self._init_plugin_class(c)
        result = dict()

        for item in working_set.iter_entry_points(group, None):
            # format: "plugin_name=plugin_module:plugin_class",
            if self.mode == MODE_EXPLICIT:
                cls = get_class(module_name=item.module_name, class_name=item.attrs[0])
                if issubclass(cls, c):
                    p = cls()
                    if self._register_plugin(result, p):
                        names = get_all_names(p)
                        for name in names:
                            result[name] = p
                        # record aliases
                        self._all_aliases.update(get_aliases(p))
            # format: "unique_string=plugin_module:superclass_name",
            elif self.mode == MODE_DYNAMIC:
                c = get_class(full_class_name=".".join(item.attrs))
                result.update(self._register_from_module(item.module_name, c))
            else:
                raise Exception("Unhandled mode: %s" % self.mode)

        return result

    def _register(self, group: str, c: Optional = None):
        """
        Registers all plugins that match the specified class.

        :param group: the entry point group to get the plugins from
        :type group: str
        :param c: the class that the plugins must be, any class derived from Plugin if None
        """
        c = self._init_plugin_class(c)

        # from entry points
        plugins = self._register_from_entry_point(group, c=c)

        # register from modules as well?
        if (len(plugins) == 0) or ((self._custom_modules is not None) and (len(self._custom_modules) > 0)) or self.has_env_modules():
            plugins.update(self._register_from_modules(c=c))

        self._plugins[group] = plugins

    def plugins(self, group: str, c: Optional = None) -> Dict[str, Plugin]:
        """
        Returns the plugins for the specified class.

        :param group: the entry point group to get the plugins for
        :type group: str
        :param c: the class that the plugins must be, any class derived from Plugin if None
        :return: the dictionary of plugins (name / plugin association)
        :rtype: dict
        """
        if group not in self._plugins:
            self._register(group, c)
        return self._plugins[group]





from ._reader import Reader
from ._writer import Writer, StreamWriter, BatchWriter
from ._filter import Filter, MultiFilter, FILTER_ACTIONS, FILTER_ACTION_KEEP, FILTER_ACTION_DISCARD
from ._execution import execute
from ._split import gcd, Splitter
from ._utils import locate_files



import importlib
import inspect
import os
import sys
import traceback
from typing import Callable, Union, List, Optional, Type, Dict

from pkg_resources import working_set

from ._types import get_class_name, get_class
from ._plugin import Plugin, get_all_names, get_aliases

DEFAULT = "DEFAULT"
""" the placeholder for the default class listers in the environment variable. """

LIST_CLASSES = "list_classes"
""" the default method for listing classes. """


def get_class_lister(class_lister: str) -> Callable:
    """
    Parses the class_lister definition and returns the function.
    The default format is "module_name:function_name".
    If ":function_name" is omitted it is assumed to be ":list_classes".

    :param class_lister: the class lister definition to parse
    :type class_lister: str
    :return: the parsed function
    """
    if ":" not in class_lister:
        module_name = class_lister
        func_name = LIST_CLASSES
    else:
        module_name, func_name = class_lister.split(":")

    try:
        module = importlib.import_module(module_name)
    except:
        raise Exception("Failed to import class lister module: %s" % module_name)

    if hasattr(module, func_name):
        func = getattr(module, func_name)
        if inspect.isfunction(func):
            return func
        else:
            raise Exception("Class lister function is not an actual function: %s" % class_lister)
    else:
        raise Exception("Class lister function '%s' not found in module '%s'!" % (func_name, module_name))


class ClassListerRegistry:
    """
    Registry for managing class hierarchies via class listers.
    """

    def __init__(self, default_class_listers: Union[str, List[str]] = None, env_class_listers: str = None,
                 excluded_class_listers: Union[str, List[str]] = None, env_excluded_class_listers: str = None):
        """

        :param default_class_listers: the default class lister(s) to use
        :type default_class_listers: str or list
        :param env_class_listers: the environment variable to retrieve the class lister(s) from
        :type env_class_listers: str
        :param excluded_class_listers: the class lister(s) to exclude from being used
        :type excluded_class_listers: str or list
        :param env_excluded_class_listers: the environmenr variable to retrieve the excluded class lister(s) from
        :type env_excluded_class_listers: str
        """

        self._classes = dict()
        self._all_aliases = set()
        self._default_class_listers = None
        self._env_class_listers = None
        self._excluded_class_listers = None
        self._env_excluded_class_listers = None
        self._custom_class_listers = None

        self.default_class_listers = default_class_listers
        self.env_class_listers = env_class_listers
        self.excluded_class_listers = excluded_class_listers
        self.env_excluded_class_listers = env_excluded_class_listers

    @property
    def default_class_listers(self) -> Optional[List[str]]:
        """
        Returns the class lister functions.

        :return: the functions
        :rtype: list
        """
        return self._default_class_listers

    @default_class_listers.setter
    def default_class_listers(self, class_listers: Optional[Union[str, List[str]]]):
        """
        Sets/unsets the class lister functions to use. Clears the class cache.

        :param class_listers: the list of class listers to use, None to unset
        :type class_listers: list
        """
        if class_listers is None:
            class_listers = ""
        if isinstance(class_listers, str):
            class_listers = [x.strip() for x in class_listers.split(",")]
        elif isinstance(class_listers, list):
            class_listers = class_listers[:]
        else:
            raise Exception("default_class_listers must be either str or list, but got: %s" % str(type(class_listers)))
        self._default_class_listers = class_listers
        self._classes = dict()

    @property
    def env_class_listers(self) -> Optional[str]:
        """
        Returns the environment variable with the class lister functions (if any).

        :return: the class lister functions, None if none set
        :rtype: str
        """
        return self._env_class_listers

    @env_class_listers.setter
    def env_class_listers(self, class_listers: Optional[str]):
        """
        Sets/unsets the environment variable with the class lister functions to use. Clears the class cache.

        :param class_listers: the environment variable with the class lister functions to use, None to unset
        :type class_listers: str
        """
        self._env_class_listers = class_listers
        self._classes = dict()

    @property
    def excluded_class_listers(self) -> Optional[List[str]]:
        """
        Returns the excluded class lister functions.

        :return: the functions
        :rtype: list
        """
        return self._excluded_class_listers

    @excluded_class_listers.setter
    def excluded_class_listers(self, excluded_class_listers: Optional[Union[str, List[str]]]):
        """
        Sets/unsets the excluded class lister functions to use. Clears the class cache.

        :param excluded_class_listers: the list of excluded class listers to use, None to unset
        :type excluded_class_listers: list
        """
        if excluded_class_listers is None:
            excluded_class_listers = ""
        if isinstance(excluded_class_listers, str):
            excluded_class_listers = [x.strip() for x in excluded_class_listers.split(",")]
        elif isinstance(excluded_class_listers, list):
            excluded_class_listers = excluded_class_listers[:]
        else:
            raise Exception("excluded_class_listers must be either str or list, but got: %s" % str(type(excluded_class_listers)))
        self._excluded_class_listers = excluded_class_listers
        self._classes = dict()

    @property
    def env_excluded_class_listers(self) -> Optional[str]:
        """
        Returns the environment variable with the excluded class lister functions (if any).

        :return: the excluded class lister functions, None if none set
        :rtype: str
        """
        return self._env_excluded_class_listers

    @env_excluded_class_listers.setter
    def env_excluded_class_listers(self, excluded_class_listers: Optional[str]):
        """
        Sets/unsets the environment variable with the excluded class lister functions to use. Clears the class cache.

        :param excluded_class_listers: the environment variable with the excluded class lister functions to use, None to unset
        :type excluded_class_listers: str
        """
        self._env_excluded_class_listers = excluded_class_listers
        self._classes = dict()

    @property
    def custom_class_listers(self) -> Optional[List[str]]:
        """
        Returns the custom class listers (if any).

        :return: the class listers, None if none set
        :rtype: list
        """
        return self._custom_class_listers

    @custom_class_listers.setter
    def custom_class_listers(self, class_listers: Optional[Union[str, List[str]]]):
        """
        Sets/unsets the custom class listers to use. Clears the plugin cache.

        :param class_listers: the list of class listers to use, None to unset
        :type class_listers: list
        """
        self._custom_class_listers = class_listers
        self._classes = dict()

    def has_env_class_listers(self) -> bool:
        """
        Checks whether an environment variable for class listers is set.

        :return: True if set
        :rtype: bool
        """
        return (self._env_class_listers is not None) \
               and (len(self._env_class_listers) > 0) \
               and (os.getenv(self._env_class_listers) is not None) \
               and (len(os.getenv(self._env_class_listers)) > 0)

    def has_env_excluded_class_listers(self) -> bool:
        """
        Checks whether an environment variable for excluded class listers is set.

        :return: True if set
        :rtype: bool
        """
        return (self._env_excluded_class_listers is not None) \
               and (len(self._env_excluded_class_listers) > 0) \
               and (os.getenv(self._env_excluded_class_listers) is not None) \
               and (len(os.getenv(self._env_excluded_class_listers)) > 0)

    def _expand_default_class_listers_placeholder(self, c: str) -> str:
        """
        Expands the DEFAULT class listers placeholder in the comma-separated class listers string.

        :param c: the class listers string to expand
        :type c: str
        :return: the expanded class listers string
        :rtype: str
        """
        if DEFAULT in c:
            if len(self.default_class_listers) > 0:
                c = c.replace(DEFAULT, ",".join(self.default_class_listers))
            else:
                c = c.replace(DEFAULT, "")
        return c

    def actual_fallback_class_listers(self) -> List[str]:
        """
        Returns list the of class listers to fall back on.
        Precedence: custom_class_listers > env_class_listers > default_class_listers

        :return: the list of class listers
        :rtype: list
        """
        if (self._custom_class_listers is not None) and (len(self._custom_class_listers) > 0):
            return self._custom_class_listers

        if self.has_env_class_listers():
            m = self._expand_default_class_listers_placeholder(os.getenv(self.env_class_listers))
            return [x.strip() for x in m.split(",")]

        return self.default_class_listers[:]

    def actual_excluded_class_listers(self) -> List[str]:
        """
        Returns list the of excluded class listers.
        Precedence: excluded_env_class_listers > excluded_class_listers

        :return: the list of class listers
        :rtype: list
        """
        if self.has_env_excluded_class_listers():
            m = self._expand_default_class_listers_placeholder(os.getenv(self._env_excluded_class_listers))
            return [x.strip() for x in m.split(",")]

        return self.excluded_class_listers[:]

    def _determine_sub_classes(self, cls: Type, module_name: str) -> List[str]:
        """
        Determines all the sub-classes of type cls in the specified module.

        :param cls: the superclass
        :param module_name: the module to look for sub-classes
        :type module_name: str
        :return: the list of sub-classes
        :rtype: list
        """
        result = []

        try:
            module = importlib.import_module(module_name)
        except:
            print("Failed to import module: %s" % module_name, file=sys.stderr)
            traceback.print_exc()
            return result

        for att_name in dir(module):
            if att_name.startswith("_"):
                continue
            if att_name.startswith("Abstract"):
                continue
            att = getattr(module, att_name)
            if inspect.isclass(att) and not inspect.isabstract(att) and issubclass(att, cls):
                try:
                    obj = att()
                except NotImplementedError:
                    pass
                except:
                    print("Problem encountered instantiating: %s" % (module_name + "." + att_name), file=sys.stderr)
                    traceback.print_exc()
                    continue
                result.append(get_class_name(att))

        return result

    def _determine_from_class_listers(self, c: str, class_listers: List[str]) -> List[str]:
        """
        Determines the derived classes via the specified class listers.

        :param c: the superclass to get the classes for
        :type c: str
        :param class_listers: the class lister functions to use
        :type class_listers: list
        :return: the determined list of subclasses
        :rtype: list
        """
        result = []

        if len(class_listers) > 0:
            cls = None

            for class_lister in class_listers:
                if class_lister == "":
                    continue

                if cls is None:
                    try:
                        cls = get_class(c)
                    except:
                        print("Failed to instantiate class: %s" % c, file=sys.stderr)
                        traceback.print_exc()
                        return result

                try:
                    func = get_class_lister(class_lister)
                except:
                    print("Problem encountered with class lister: %s" % class_lister, file=sys.stderr)
                    traceback.print_exc()
                    continue

                if self.excluded_class_listers is not None:
                    if class_lister in self.excluded_class_listers:
                        continue

                if inspect.isfunction(func):
                    class_dict = func()
                    if c in class_dict:
                        for sub_module in class_dict[c]:
                            sub_classes = self._determine_sub_classes(cls, sub_module)
                            result.extend(sub_classes)

        return result

    def _determine_from_entry_points(self, c: str) -> List[str]:
        """
        Determines the derived classes via class listers defined as entry points.

        :param c: the superclass to get the classes for
        :type c: str
        :return: the determined list of subclasses
        :rtype: list
        """
        result = []
        class_listers = []
        for item in working_set.iter_entry_points("class_lister", None):
            if len(item.attrs) > 0:
                # format: "name=module:function"
                class_listers.append(item.module_name + ":" + item.attrs[0])
            else:
                # format: "name=module"
                class_listers.append(item.module_name)
        if len(class_listers) > 0:
            result = self._determine_from_class_listers(c, class_listers)
        return result

    def _determine_from_env(self, c: str) -> List[str]:
        """
        Determines the derived classes via class listers defined through the environment variable.

        :param c: the superclass to get the classes for
        :type c: str
        :return: the determined list of subclasses
        :rtype: list
        """
        result = []

        if os.getenv(self.env_class_listers) is not None:
            # format: "classlister1,classlister2,..."
            # classlister format: "module_name:function_name" or "module_name" if "list_classes" as method
            class_listers = os.getenv(self.env_class_listers).split(",")
            result = self._determine_from_class_listers(c, class_listers)

        return result

    def _initialize(self, c: str):
        """
        Initializes the class cache for the specified superclass.

        :param c: the superclass to initialize the cache for
        :type c: str
        :return: the list of classes for the superclass
        :rtype: list
        """
        all_classes = set()

        # from entry points
        entry_point_classes = self._determine_from_entry_points(c)
        if entry_point_classes is not None:
            all_classes.update(entry_point_classes)

        # register from class listers as well?
        if (len(all_classes) == 0) or ((self._custom_class_listers is not None) and (len(self._custom_class_listers) > 0)) or self.has_env_class_listers():
            actual = self.actual_fallback_class_listers()
            all_classes.update(self._determine_from_class_listers(c, actual))

        # excluded classes?
        excluded_listers = self.actual_excluded_class_listers()
        excluded_classes = self._determine_from_class_listers(c, excluded_listers)
        for cls in excluded_classes:
            if cls in all_classes:
                all_classes.remove(cls)

        self._classes[c] = sorted(list(all_classes))


    def plugins(self, c: Union[str, Type], fail_if_empty: bool = True) -> Dict[str, Plugin]:
        """
        Returns the classes for the specified superclass.

        :param c: the super class to get the derived classes for (classname or type)
        :param fail_if_empty: whether to raise an exception if no classes present
        :type fail_if_empty: bool
        :return: the list of classes
        :rtype: list
        """
        if not isinstance(c, str):
            c = get_class_name(c)
        if c not in self._classes:
            self._initialize(c)
        result = dict()
        if c in self._classes:
            for cname in self._classes[c]:
                try:
                    cls = get_class(cname)
                    plugin = cls()
                    names = get_all_names(plugin)
                    for name in names:
                        result[name] = plugin
                        # record any aliases
                        self._all_aliases.update(get_aliases(plugin))
                except NotImplementedError:
                    pass
                except:
                    print("Failed to instantiate: %s" % cname)
                    traceback.print_exc()
        if fail_if_empty and (len(result) == 0):
            raise Exception("No classes found for: %s" % c)
        return result

    @property
    def all_aliases(self) -> List[str]:
        """
        Returns a sorted list of all known aliases.
        Due to dynamic instantiation the call to all_aliases must come after the relevant plugins(...) call.

        :return: the list of all known aliases
        :rtype: list
        """
        return sorted(list(self._all_aliases))

    def is_alias(self, name: str) -> bool:
        """
        Checks whether the plugin name is an alias.
        Due to dynamic instantiation the call to is_alias(...) must come after the relevant plugins(...) call.

        :param name: the plugin name to check
        :type name: str
        :return: True if an alias
        :rtype: bool
        """
        return name in self._all_aliases



import os
from typing import List

from ._plugin import Plugin, get_aliases
from ._placeholders import PlaceholderSupporter, placeholder_help

HELP_FORMAT_TEXT = "text"
HELP_FORMAT_MARKDOWN = "markdown"
HELP_FORMATS = [
    HELP_FORMAT_TEXT,
    HELP_FORMAT_MARKDOWN,
]


def generate_plugin_usage(plugin: Plugin, help_format: str = HELP_FORMAT_TEXT, heading_level: int = 1):
    """
    Generates the usage help screen for the specified plugin.

    :param plugin: the plugin to generate the usage for (name used on command-line)
    :type plugin: Plugin
    :param help_format: the format to use for the output
    :type help_format: str
    :param heading_level: the level to use for the heading (markdown)
    :type heading_level: int
    """
    if help_format not in HELP_FORMATS:
        raise Exception("Unhandled help format: %s" % help_format)

    aliases = get_aliases(plugin)
    result = ""
    if help_format == HELP_FORMAT_TEXT:
        result += "\n" + plugin.name() + "\n" + "=" * len(plugin.name()) + "\n"
        result = result.strip()
        result += "\n\n"
        if len(aliases) > 0:
            result += "Alias(es): %s\n\n" % ", ".join(aliases)
        result += plugin.format_help() + "\n"
        if isinstance(plugin, PlaceholderSupporter):
            result += "\n" + placeholder_help(markdown=False, obj=plugin) + "\n"
    elif help_format == HELP_FORMAT_MARKDOWN:
        result += "#"*heading_level + " " + plugin.name() + "\n"
        result += "\n"
        result = result.strip()
        result += "\n\n"
        if len(aliases) > 0:
            result += "Alias(es): %s\n\n" % ", ".join(aliases)
        result += plugin.description() + "\n"
        result += "\n"
        result += "```\n"
        result += plugin.format_help()
        result += "```\n"
        if isinstance(plugin, PlaceholderSupporter):
            result += "\n" + placeholder_help(markdown=True, obj=plugin) + "\n"
    else:
        raise Exception("Unhandled help format: %s" % help_format)

    return result


def generate_help(plugins: List[Plugin], help_format: str = HELP_FORMAT_TEXT, heading_level: int = 1,
                  output_path: str = None):
    """
    Generates and outputs the help screen for the plugin.

    :param plugins: the plugins to generate the help screens for
    :type plugins: list
    :param help_format: the format to output
    :type help_format: str
    :param heading_level: the heading level to use (markdown)
    :type heading_level: int
    :param output_path: the dir to save the output to, uses stdout if None
    :type output_path: str
    """
    if help_format not in HELP_FORMATS:
        raise Exception("Unknown help format: %s" % help_format)

    for p in plugins:
        help_screen = generate_plugin_usage(p, help_format=help_format, heading_level=heading_level)
    
        if output_path is None:
            print(help_screen)
        else:
            if help_format == HELP_FORMAT_TEXT:
                suffix = ".txt"
            elif help_format == HELP_FORMAT_MARKDOWN:
                suffix = ".md"
            else:
                raise Exception("Unhandled help format: %s" % help_format)
    
            if os.path.isdir(output_path):
                output_file = os.path.join(output_path, p.name() + suffix)
            else:
                output_file = output_path
            with open(output_file, "w") as fp:
                fp.write(help_screen)



import abc
from typing import List

from wai.logging import LOGGING_WARNING
from seppl import InputConsumer, OutputProducer, PluginWithLogging, Initializable, init_initializable, Session, SessionHandler

FILTER_ACTION_KEEP = "keep"
FILTER_ACTION_DISCARD = "discard"
FILTER_ACTIONS = [FILTER_ACTION_KEEP, FILTER_ACTION_DISCARD]


class Filter(PluginWithLogging, InputConsumer, OutputProducer, SessionHandler, Initializable, abc.ABC):
    """
    Base class for filters.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the handler.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self._session = None
        self._last_input = None

    @property
    def session(self) -> Session:
        """
        Returns the current session object

        :return: the session object
        :rtype: Session
        """
        return self._session

    @session.setter
    def session(self, s: Session):
        """
        Sets the session object to use.

        :param s: the session object
        :type s: Session
        """
        self._session = s

    def _has_input_changed(self, current_input: str = None, update: bool = False) -> bool:
        """
        Checks whether the current input is different from the last one we processed.

        :param current_input: the current input, uses the current_input from the session if None
        :type current_input: str
        :param update: whether to update the last input immediately
        :type update: bool
        :return: True if input has changed
        :rtype: bool
        """
        if current_input is None:
            current_input = self.session.current_input
        result = self._last_input != current_input
        if update:
            self._update_last_input(current_input)
        return result

    def _update_last_input(self, current_input: str):
        """
        Updates the last input that was processed.

        :param current_input: the "new" last input
        :type current_input: str
        """
        self._last_input = current_input

    def _requires_list_input(self) -> bool:
        """
        Returns whether lists are expected as input for the _process method.

        :return: True if list inputs are expected by the filter
        :rtype: bool
        """
        return False

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        raise NotImplementedError()

    def process(self, data):
        """
        Processes the data record.

        :param data: the record(s) to process
        :return: the potentially updated record or None if to drop
        """
        if isinstance(data, list):
            if self._requires_list_input():
                result = self._do_process(data)
            else:
                result = []
                for d in data:
                    r = self._do_process(d)
                    if r is not None:
                        if isinstance(r, list):
                            result.extend(r)
                        else:
                            result.append(r)
                if len(result) == 1:
                    result = result[0]
        else:
            if self._requires_list_input():
                result = self._do_process([data])
            else:
                result = self._do_process(data)
        return result


class MultiFilter(Filter, Initializable):
    """
    Combines multiple filters.
    """

    def __init__(self, filters: List[Filter] = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initialize with the specified filters.

        :param filters: the filters to use
        :type filters: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.filters = None if (filters is None) else filters[:]

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "multi-filter"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Combines multiple filters."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        if (self.filters is not None) and (len(self.filters) > 0):
            return self.filters[0].accepts()
        else:
            return list()

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        if (self.filters is not None) and (len(self.filters) > 0):
            return self.filters[-1].generates()
        else:
            return list()

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        for f in self.filters:
            f.session = self.session
            init_initializable(f, "filter", raise_again=True)

    def _requires_list_input(self) -> bool:
        """
        Returns whether lists are supported as input.

        :return: True if list inputs are natively handled by the filter
        :rtype: bool
        """
        return True

    def _do_process(self, data):
        """
        Processes the data record.

        :param data: the record to process
        :return: the potentially updated record or None if to drop
        """
        result = data
        for f in self.filters:
            result = f.process(result)
            if result is None:
                break
        return result

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        for f in self.filters:
            f.finalize()



import copy
import shlex

from typing import List, Dict, Tuple, Iterable, Set, Optional
from ._plugin import Plugin


def escape_args(args: List[str]) -> List[str]:
    """
    Escapes any unicode characters in the arguments.

    :param args: the arguments to process
    :type args: list
    :return: the (potentially) updated arguments
    :rtype: list
    """
    result = []
    for arg in args:
        result.append(arg.encode("unicode_escape").decode())
    return result


def unescape_args(args: List[str]) -> List[str]:
    """
    Unescapes unicode characters in the arguments.

    :param args: the arguments to process
    :type args: list
    :return: the (potentially) updated arguments
    :rtype: list
    """
    result = []
    for arg in args:
        result.append(arg.encode().decode("unicode_escape"))
    return result


def split_cmdline(cmdline: str, unescape: bool = False) -> List[str]:
    """
    Splits the command-line into arguments.

    :param cmdline: the commandline to split
    :type cmdline: str
    :param unescape: whether to unescape unicode chars
    :type unescape: bool
    :return: the list of arguments
    :rtype: list
    """
    result = shlex.split(cmdline)
    if unescape:
        result = unescape_args(result)
    return result

def resolve_handler(search: str, handlers: Set[str]) -> Optional[str]:
    """
    Tries to find the "search" string among the handlers, exact and partial match.

    :param search: the potential handler to find
    :type search: str
    :param handlers: the set of valid handlers to match against
    :type handlers: set
    :return: the match or None if failed to find
    :rtype: str or None
    """
    # exact match?
    if search in handlers:
        return search
    # unique partial match?
    matches = []
    for handler in handlers:
        if handler.startswith(search):
            matches.append(handler)
    if len(matches) == 1:
        return matches[0]
    # nothing found
    return None


def split_args(args: List[str], handlers: List[str], unescape: bool = False) -> Dict[str, List[str]]:
    """
    Splits the command-line arguments into handler and their associated arguments.
    Special entry "" is used for global options.

    :param args: the command-line arguments to split
    :type args: list
    :param handlers: the list of valid handler names
    :type handlers: list
    :param unescape: whether to unescape unicode chars
    :type unescape: bool
    :return: the dictionary for handler index / handler name + options list
    :rtype: dict
    """
    handlers_set = set(handlers)
    result = dict()
    last_handler = ""
    last_args = []

    if unescape:
        args = unescape_args(args)

    for arg in args:
        handler = resolve_handler(arg, handlers_set)
        if handler is not None:
            if len(last_handler) > 0:
                result[str(len(result))] = last_args
            else:
                result[""] = last_args
            last_handler = handler
            last_args = [handler]
            continue
        else:
            last_args.append(arg)

    if len(last_args) > 0:
        result[str(len(result))] = last_args

    return result


def args_to_objects(args: Dict[str, List[str]], valid_plugins: Dict[str, Plugin], allow_global_options: bool = False, allow_unknown_args: bool = False, unescape: bool = False) -> List[Plugin]:
    """
    Instantiates the plugins from the parsed arguments dictionary.

    :param args: the arguments dictionary generated by split_args
    :type args: dict
    :param valid_plugins: the dictionary of valid plugins to use as templates
    :type valid_plugins: dict
    :param allow_global_options: whether global options are allowed (ie options that don't follow a plugin name)
    :type allow_global_options: bool
    :param allow_unknown_args: whether to allow unknown args (eg typos or unknown plugins)
    :type allow_unknown_args: bool
    :param unescape: whether to unescape unicode chars
    :type unescape: bool
    :return: the list of instantiated plugins
    :rtype: list
    """
    result = []
    for key in args:
        if key == "":
            if (not allow_global_options) and (len(args[""]) > 0):
                raise Exception("No global options allowed (found: %s)!" % str(args[""]))
            else:
                continue

        name = args[key][0]
        plugin = copy.deepcopy(valid_plugins[name])
        sub_args = args[key][1:]
        if unescape:
            sub_args = unescape_args(sub_args)
        unknown = plugin.parse_args(sub_args)
        if not allow_unknown_args and (len(unknown) > 0):
            raise Exception("Found unknown argument(s) for plugin '%s': %s" % (plugin.name(), str(unknown)))
        result.append(plugin)
    return result


def is_help_requested(args: List[str]) -> Tuple[bool, bool, str]:
    """
    Checks whether help was requested.

    :param args: the arguments to check
    :type args: list
    :return: the tuple of help requested: (help_requested, plugin_details, plugin_name)
    :rtype: tuple
    """
    help_requested = False
    plugin_details = False
    plugin_name = None
    for index, arg in enumerate(args):
        if (arg == "-h") or (arg == "--help"):
            help_requested = True
            break
        if arg == "--help-all":
            help_requested = True
            plugin_details = True
            break
        if arg == "--help-plugin":
            help_requested = True
            if index < len(args) - 1:
                plugin_name = args[index + 1]
            break
    return help_requested, plugin_details, plugin_name


def enumerate_plugins(plugins: Iterable[str], aliases: List[str] = None, alias_flag: str = "*", prefix: str = "", width: int = 72) -> str:
    """
    Turns the list of plugin names into a string.

    :param plugins: the plugin names to turn into a string
    :type plugins: Iterable
    :param aliases: the list of known aliases (for flagging them in the generated list)
    :type aliases: list
    :param alias_flag: the string to use for identifying aliases
    :type alias_flag: str
    :param prefix: the prefix string to use for each line
    :type prefix: str
    :param width: the maximum width of the string before adding a newline
    :type width: int
    :return: the generated string
    :rtype: str
    """
    result = []
    line = prefix
    for plugin in sorted(plugins):
        if (len(line) > 0) and (line[-1] != " "):
            line += ", "
        if len(line) + len(plugin) >= width:
            result.append(line)
            line = prefix + plugin
        else:
            line += plugin
        if aliases is not None:
            if plugin in aliases:
                line += alias_flag

    if len(line) > 0:
        result.append(line)
    return "\n".join(result)



import shlex
import sys
import traceback

from seppl import unescape_args


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    if args is None:
        args = sys.argv[1:]
    unescaped = unescape_args(args)
    print(shlex.join(unescaped))


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        print("options: %s" % str(sys.argv[1:]), file=sys.stderr)
        return 1


if __name__ == '__main__':
    main()



import abc
from typing import Iterable

from wai.logging import LOGGING_WARNING

from seppl import OutputProducer, PluginWithLogging
from seppl import SessionHandler, Session


class Reader(PluginWithLogging, OutputProducer, SessionHandler, abc.ABC):
    """
    Ancestor of classes that read data.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the handler.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self._session = None

    @property
    def session(self) -> Session:
        """
        Returns the current session object

        :return: the session object
        :rtype: Session
        """
        return self._session

    @session.setter
    def session(self, s: Session):
        """
        Sets the session object to use.

        :param s: the session object
        :type s: Session
        """
        self._session = s

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        raise NotImplementedError()

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        raise NotImplementedError()



import glob
import os

from typing import Union, List
from seppl import expand_placeholders


def locate_files(inputs: Union[str, List[str]], input_lists: Union[str, List[str]] = None,
                 recursive: bool = False, fail_if_empty: bool = False, default_glob: str = None) -> List[str]:
    """
    Locates all the files from the specified inputs, which may contain globs.
    glob results get sorted to ensure the same file order each time.
    If default_glob is not None and the inputs are pointing to directories, then default_glob
    gets appended. Automatically expands placeholders that are not input-based, including
    inside the input list text files.

    :param inputs: the input path(s) with optional globs
    :type inputs: str or list
    :param input_lists: text file(s) that list the actual input files to use
    :type input_lists: str or list
    :param recursive: for supporting recursive globs
    :type recursive: bool
    :param fail_if_empty: whether to throw an exception if no files were located
    :type fail_if_empty: bool
    :param default_glob: the default glob to use, ignored if None
    :type default_glob: str
    :return: the expanded list of files
    :rtype: list
    """
    if (inputs is None) and (input_lists is None):
        raise Exception("Neither input paths nor input lists provided!")

    if inputs is not None:
        if isinstance(inputs, str):
            inputs = [inputs]
        elif isinstance(inputs, list):
            inputs = inputs
        else:
            raise Exception("Invalid inputs, must be string(s)!")

        # expand placeholders
        inputs = [expand_placeholders(x) for x in inputs]

        if default_glob is not None:
            for i, inp in enumerate(inputs):
                if os.path.isdir(inp):
                    inputs[i] = os.path.join(inp, default_glob)

    if input_lists is not None:
        if isinstance(input_lists, str):
            input_lists = [input_lists]
        elif isinstance(input_lists, list):
            input_lists = input_lists
        else:
            raise Exception("Invalid input lists, must be string(s)!")

        # expand placeholders
        input_lists = [expand_placeholders(x) for x in input_lists]

    result = []

    # globs
    if inputs is not None:
        for inp in inputs:
            for f in sorted(glob.glob(inp, recursive=recursive)):
                if os.path.isdir(f):
                    continue
                result.append(f)

    # path lists
    if input_lists is not None:
        for inp in input_lists:
            if not os.path.exists(inp):
                print("WARNING: Input list does not exist: %s" % inp)
                continue
            if os.path.isdir(inp):
                print("WARNING: Input list points to directory: %s" % inp)
                continue
            with open(inp, "r") as fp:
                lines = [expand_placeholders(x.strip()) for x in fp.readlines()]
            for line in lines:
                if len(line) == 0:
                    continue
                if not os.path.exists(line):
                    print("WARNING: Path from input list '%s' does not exist: %s" % (inp, line))
                    continue
                result.append(line)

    if fail_if_empty and (len(result) == 0):
        raise Exception("Failed to locate any files using: %s" % str(inputs))

    return result



import argparse
import logging
from dataclasses import dataclass

from wai.logging import set_logging_level

from ._placeholders import expand_placeholders


@dataclass
class Session:
    """
    Session object shared among reader, filter(s), writer.
    """
    options: argparse.Namespace = None
    """ global options. """

    logger: logging.Logger = None
    """ the global logger. """

    count: int = 0
    """ the record counter. """

    current_input = None
    """ the current input etc. """

    def _add_option(self, name: str, value):
        """
        Adds the key/value to the global options.

        :param name: the name of the option
        :type name: str
        :param value: the value of the option
        """
        if self.options is None:
            self.options = argparse.Namespace()
        setattr(self.options, name, value)

    def set_logging_level(self, level: str):
        """
        Sets the global logging level.

        :param level: the level
        :type level: str
        :return: itself
        :rtype: Session
        """
        self._add_option("logging_level", level)
        set_logging_level(self.logger, level)
        return self

    def expand_placeholders(self, template: str) -> str:
        """
        Expands the placeholder in the template using the current input.

        :param template: the template to expand
        :type template: str
        :return: the expanded string
        :rtype: str
        """
        return expand_placeholders(template, current_input=self.current_input)


class SessionHandler(object):
    """
    Mixin for classes that support session objects.
    """

    @property
    def session(self) -> Session:
        """
        Returns the current session object

        :return: the session object
        :rtype: Session
        """
        raise NotImplementedError()

    @session.setter
    def session(self, s: Session):
        """
        Sets the session object to use.

        :param s: the session object
        :type s: Session
        """
        raise NotImplementedError()



import importlib
import json

from typing import List, Dict
from ._plugin import Plugin


def _to_entry_point(plugin: Plugin) -> str:
    """
    Turns the plugin into an entry point.

    :param plugin: the object to convert
    :type plugin: CommandlineHandler
    :return: the generated entry point
    :rtype: str
    """
    m = plugin.__module__

    # can we hide a private module?
    parts = m.split(".")
    if parts[-1].startswith("_"):
        parts.pop()
        m = ".".join(parts)
        try:
            importlib.import_module(m)
        except:
            # can't remove the last and private module, so we'll stick with the full path
            m = plugin.__module__

    result = plugin.name() + "=" + m + ":" + plugin.__class__.__name__
    return result


def _to_entry_points(plugins: List[Plugin]) -> List[str]:
    """
    Turns the plugins into a list of entry points.

    :param plugins: the plugins to convert
    :type plugins: list
    :return: the list of entry points
    :rtype: list
    """
    result = list()
    for plugin in plugins:
        result.append(_to_entry_point(plugin))
    return result


def generate_entry_points(entry_points: Dict[str, List[Plugin]]) -> str:
    """
    Generates the entry points string.
    Automatically suppresses empty entry point groups with an empty list of plugins.

    :param entry_points: the plugin lists, one list per entry_points group
    :type entry_points: dict
    """
    result = dict()
    keys = list(entry_points.keys())
    for k in keys:
        if len(entry_points[k]) == 0:
            entry_points.pop(k)
        result[k] = _to_entry_points(entry_points[k])

    return "entry_points=" + json.dumps(result, indent=4)



import traceback
from typing import Union, List, Optional

from seppl import init_initializable, Initializable, Session
from ._reader import Reader
from ._filter import Filter, MultiFilter
from ._writer import Writer, StreamWriter, BatchWriter


def _stream_execution(reader: Reader, filter_: Optional[Filter], writer: Optional[Writer], session: Session):
    """
    Executes the pipeline in streaming mode.

    :param reader: the reader to use
    :type reader: Reader
    :param filter_: the filter to use
    :type filter_: list or Filter
    :param writer: the writer to use
    :type writer: Writer
    :param session: the session object to use
    :type session: Session
    """
    if writer is not None:
        if not isinstance(writer, StreamWriter):
            raise Exception("Not a stream writer: %s" % str(type(writer)))

    while not reader.has_finished():
        for item in reader.read():
            if item is None:
                continue
            session.count += 1
            if (filter_ is not None) and (item is not None):
                item = filter_.process(item)
            if item is not None:
                if writer is not None:
                    writer.write_stream(item)
            if session.count % session.options.update_interval == 0:
                session.logger.info("%d records processed..." % session.count)


def _batch_execution(reader: Reader, filter_: Optional[Filter], writer: Optional[Writer], session: Session):
    """
    Executes the pipeline in batch mode.

    :param reader: the reader to use
    :type reader: Reader
    :param filter_: the filter(s) to use, can be None
    :type filter_: list or Filter
    :param writer: the writer to use
    :type writer: Writer
    :param session: the session object to use
    :type session: Session
    """
    data = []
    while not reader.has_finished():
        for item in reader.read():
            if item is None:
                continue
            session.count += 1
            data.append(item)
            if session.count % session.options.update_interval == 0:
                session.logger.info("%d records read..." % session.count)

    if filter_ is not None:
        data = filter_.process(data)
        session.logger.info("%d records filtered..." % session.count)
        if not isinstance(data, list):
            data = [data]

    if writer is not None:
        if isinstance(writer, StreamWriter):
            count = 0
            for item in data:
                count += 1
                writer.write_stream(item)
                if count % session.options.update_interval == 0:
                    session.logger.info("%d records written..." % count)
        elif isinstance(writer, BatchWriter):
            writer.write_batch(data)
        else:
            raise Exception("Neither stream nor batch writer: %s" % str(type(writer)))


def execute(reader: Reader, filters: Optional[Union[Filter, List[Filter]]], writer: Writer,
            session: Session = None):
    """
    Executes the pipeline.

    :param reader: the reader to use
    :type reader: Reader
    :param filters: the filter(s) to use, can be None
    :type filters: list or Filter
    :param writer: the writer to use
    :type writer: Writer
    :param session: the session object to use, creates default one if None
    :type session: Session
    """
    # assemble filter
    if isinstance(filters, Filter):
        filter_ = filters
    elif isinstance(filters, list):
        filter_ = MultiFilter(filters=filters)
    elif filters is None:
        filter_ = None
    else:
        raise Exception("Unhandled filter(s) type: %s" % str(type(filters)))

    # propagate session
    if session is None:
        session = Session()
    reader.session = session
    if filter_ is not None:
        filter_.session = session
    if writer is not None:
        writer.session = session

    # initialize
    if isinstance(reader, Initializable) and not init_initializable(reader, "reader"):
        return
    if (filter_ is not None) and isinstance(filter_, Initializable) and not init_initializable(filter_, "filter"):
        return
    if (writer is not None) and isinstance(writer, Initializable) and not init_initializable(writer, "writer"):
        return

    # process data
    try:
        if session.options.force_batch or isinstance(writer, BatchWriter):
            _batch_execution(reader, filter_, writer, session)
        else:
            _stream_execution(reader, filter_, writer, session)
        session.logger.info("%d records processed in total." % session.count)
    except:
        traceback.print_exc()

    # clean up
    if isinstance(reader, Initializable):
        reader.finalize()
    if (filter_ is not None) and isinstance(filter_, Initializable):
        filter_.finalize()
    if (writer is not None) and isinstance(writer, Initializable):
        writer.finalize()



from typing import Optional, Dict


class MetaDataHandler(object):
    """
    Mixin for classes that manage meta-data.
    """

    def has_metadata(self) -> bool:
        """
        Returns whether meta-data is present.

        :return: True if meta-data present
        :rtype: bool
        """
        raise NotImplementedError()

    def get_metadata(self) -> Optional[Dict]:
        """
        Returns the meta-data.

        :return: the meta-data, None if not available
        :rtype: dict
        """
        raise NotImplementedError()

    def set_metadata(self, metadata: Optional[Dict]):
        """
        Sets the meta-data to use.

        :param metadata: the new meta-data, can be None
        :type metadata: dict
        """
        raise NotImplementedError()


def get_metadata(o) -> Optional[Dict]:
    """
    Retrieves the meta-data from the specified object.

    :param o: the object to get the meta-data from
    :return: the meta-data, None if not available
    """
    if isinstance(o, MetaDataHandler):
        return o.get_metadata()
    if hasattr(o, "meta"):
        obj = getattr(o, "meta")
        if isinstance(obj, dict):
            return obj
    return None


def add_metadata(meta: Optional[Dict], key: str, value) -> Dict:
    """
    Adds the specified key/value pair to the meta-data.
    If the provided meta-data dictionary is empty, it gets instantiated.
    If value is None, nothing gets added.
    If value is a string and empty, nothing gets added.

    :param meta: the meta-data to add to
    :type meta: dict
    :param key: the key to store the value under
    :type key: str
    :param value: the value to store
    :return: the created or updated dictionary
    :rtype: dict
    """
    if value is None:
        return meta
    if isinstance(value, str) and (len(value) == 0):
        return meta
    if meta is None:
        meta = dict()
    meta[key] = value
    return meta



import os
import tempfile

from typing import List


PH_HOME = "{HOME}"
PH_CWD = "{CWD}"
PH_TMP = "{TMP}"
PH_INPUT_PATH = "{INPUT_PATH}"
PH_INPUT_NAMEEXT = "{INPUT_NAMEEXT}"
PH_INPUT_NAMENOEXT = "{INPUT_NAMENOEXT}"
PH_INPUT_EXT = "{INPUT_EXT}"
PH_INPUT_PARENT_PATH = "{INPUT_PARENT_PATH}"
PH_INPUT_PARENT_NAME = "{INPUT_PARENT_NAME}"
PLACEHOLDERS = [
    PH_HOME,
    PH_CWD,
    PH_TMP,
    PH_INPUT_PATH,
    PH_INPUT_NAMEEXT,
    PH_INPUT_NAMENOEXT,
    PH_INPUT_EXT,
    PH_INPUT_PARENT_PATH,
    PH_INPUT_PARENT_NAME,
]
PLACEHOLDERS_INPUT_BASED = {
    PH_HOME: False,
    PH_CWD: False,
    PH_TMP: False,
    PH_INPUT_PATH: True,
    PH_INPUT_NAMEEXT: True,
    PH_INPUT_NAMENOEXT: True,
    PH_INPUT_EXT: True,
    PH_INPUT_PARENT_PATH: True,
    PH_INPUT_PARENT_NAME: True,
}
PLACEHOLDERS_LAMBDA = {
    PH_HOME: lambda i: os.path.expanduser("~"),
    PH_CWD: lambda i: os.getcwd(),
    PH_TMP: lambda i: tempfile.gettempdir(),
    PH_INPUT_PATH: lambda i: os.path.dirname(i),
    PH_INPUT_NAMEEXT: lambda i: os.path.basename(i),
    PH_INPUT_NAMENOEXT: lambda i: os.path.splitext(os.path.basename(i))[0],
    PH_INPUT_EXT: lambda i: os.path.splitext(i)[1],
    PH_INPUT_PARENT_PATH: lambda i: os.path.dirname(os.path.dirname(i)) if (len(os.path.dirname(i)) > 0) else "",
    PH_INPUT_PARENT_NAME: lambda i: os.path.basename(os.path.dirname(i)) if (len(os.path.dirname(i)) > 0) else "",
}
PLACEHOLDERS_DESCRIPTION = {
    PH_HOME: "The home directory of the current user.",
    PH_CWD: "The current working directory.",
    PH_TMP: "The temp directory.",
    PH_INPUT_PATH: "The directory part of the current input, i.e., '/some/where' of input '/some/where/file.txt'.",
    PH_INPUT_NAMEEXT: "The name (incl extension) of the current input, i.e., 'file.txt' of input '/some/where/file.txt'.",
    PH_INPUT_NAMENOEXT: "The name (excl extension) of the current input, i.e., 'file' of input '/some/where/file.txt'.",
    PH_INPUT_EXT: "The extension of the current input (incl dot), i.e., '.txt' of input '/some/where/file.txt'.",
    PH_INPUT_PARENT_PATH: "The directory part of the parent directory of the current input, i.e., '/some' of input '/some/where/file.txt'.",
    PH_INPUT_PARENT_NAME: "The name of the parent directory of the current input, i.e., 'where' of input '/some/where/file.txt'.",
}
USER_DEFINED_PLACEHOLDERS = set()


class PlaceholderSupporter:
    """
    Indicator mixin whether a class supports placeholders in some form.
    Used for outputting help information.
    """
    pass


class InputBasedPlaceholderSupporter(PlaceholderSupporter):
    """
    Indicator mixin whether a class supports input-based placeholders.
    Used for outputting help information.
    """
    pass


def add_placeholder(placeholder: str, description: str, input_based: bool, lambda_func):
    """
    Allows adding a custom placeholder.

    :param placeholder: the placeholder itself
    :type placeholder: str
    :param description: the description for the placeholder, used in the placeholder_help() method
    :type description: str
    :param input_based: whether the placeholder relies on the current input
    :type input_based: str
    :param lambda_func: the lambda to use for expanding the placeholder, takes one argument: current input
    """
    if "{" not in placeholder:
        placeholder = "{" + placeholder + "}"
    if placeholder not in PLACEHOLDERS:
        PLACEHOLDERS.append(placeholder)
    PLACEHOLDERS_DESCRIPTION[placeholder] = description
    PLACEHOLDERS_INPUT_BASED[placeholder] = input_based
    PLACEHOLDERS_LAMBDA[placeholder] = lambda_func


def expand_placeholders(template: str, current_input: str = None) -> str:
    """
    Expands the placeholder in the template using the current input.

    :param current_input: the current input dir/file to use for the expansion
    :type current_input: str
    :param template: the template to expand
    :type template: str
    :return: the expanded string
    :rtype: str
    """
    result = template

    if "{" in result:
        for ph in PLACEHOLDERS:
            input_based = PLACEHOLDERS_INPUT_BASED[ph]
            lambda_func = PLACEHOLDERS_LAMBDA[ph]
            value = None
            if not input_based:
                value = lambda_func(None)
            elif input_based and (current_input is not None):
                value = lambda_func(current_input)
            if value is not None:
                result = result.replace(ph, value)

    return result


def placeholders(input_based: bool = False) -> List[str]:
    """
    Returns the placeholder names as list. Excludes user-defined placeholders.

    :param input_based: whether to include input based ones
    :type input_based: bool
    :return: the list of placeholders
    :rtype: list
    """
    result = []
    for ph in PLACEHOLDERS_INPUT_BASED:
        if ph in USER_DEFINED_PLACEHOLDERS:
            continue
        if (not input_based) and (not PLACEHOLDERS_INPUT_BASED[ph]):
            result.append(ph)
        elif input_based and PLACEHOLDERS_INPUT_BASED[ph]:
            result.append(ph)
    return result


def placeholder_list(input_based: bool = False, obj=None) -> str:
    """
    Returns a short string of supported placeholders as list, e.g., to be used in the help string
    of argparse options, e.g.: placeholder_list(obj=self).

    :param input_based: whether to include input based ones
    :type input_based: bool
    :param obj: the object to determine the placeholder mixin from, overrides input_based parameter
    :return: the generated string
    :rtype: str
    """
    if obj is not None:
        input_based = isinstance(obj, InputBasedPlaceholderSupporter)
    return "Supported placeholders: %s" % ", ".join(placeholders(input_based=input_based))


def placeholder_help(input_based: bool = False, obj=None, markdown: bool = False) -> str:
    """
    Returns help on placeholders.

    :param input_based: whether to include input based ones
    :type input_based: bool
    :param obj: the object to determine the placeholder mixin from, overrides input_based parameter
    :param markdown: whether to generate markdown or plain text
    :type markdown: bool
    :return: the generated help string
    :rtype: str
    """
    if obj is not None:
        input_based = isinstance(obj, InputBasedPlaceholderSupporter)
    result = "Available placeholders:"
    if markdown:
        result += "\n"
    for ph in placeholders(input_based=input_based):
        if markdown:
            result += "\n* `" + ph + "`: " + PLACEHOLDERS_DESCRIPTION[ph].replace("'", "`")
        else:
            result += "\n- " + ph + ": " + PLACEHOLDERS_DESCRIPTION[ph]
    return result


def load_user_defined_placeholders(path: str):
    """
    Loads placeholders from the specified text file (format: key=value) that are not input-based.
    With "key" being the name of the placeholder without the curly brackets and value the path
    that it represents.

    :param path: the file with the placeholders to load
    :type path: str
    """
    with open(path, "r") as fp:
        lines = fp.readlines()
    for line in lines:
        line = line.strip()
        if len(line) == 0:
            continue
        parts = line.split("=")
        if len(parts) == 2:
            ph = "{" + parts[0] + "}"
            USER_DEFINED_PLACEHOLDERS.add(ph)
            add_placeholder(ph, "", False, lambda i: parts[1])
        else:
            print("Invalid placeholder format (key=value): %s" % line)



import argparse
import logging
import sys
import traceback

from typing import List
from ._types import classes_to_str, get_class_name, AnyData

from wai.logging import LOGGING_WARNING, set_logging_level, add_logging_level, add_logger_name


UNKNOWN_ARGS = "unknown_args"


class Plugin:
    """
    Base class for plugins.
    """

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        raise NotImplementedError()

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        raise NotImplementedError()

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = argparse.ArgumentParser(
            description=self.description(),
            prog=self.name(),
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument(UNKNOWN_ARGS, nargs=argparse.REMAINDER, help=argparse.SUPPRESS)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        pass

    def parse_args(self, args: List[str]) -> list:
        """
        Parses the command-line arguments.

        :param args: the arguments to parse
        :type args: list
        :return: any unknown args
        :rtype: list
        """
        parser = self._create_argparser()
        ns = parser.parse_args(args)
        self._apply_args(ns)
        return ns.unknown_args

    def print_help(self):
        """
        Outputs the help in the console.
        """
        self._create_argparser().print_help()

    def format_help(self) -> str:
        """
        Returns the formatted help string.

        :return: the help string
        :rtype: str
        """
        return self._create_argparser().format_help()


class AliasSupporter(object):
    """
    Mixin for classes that support alias names.
    """

    def aliases(self) -> List[str]:
        """
        Returns the aliases under which the plugin is known under/available as well.

        :return: the aliases
        :rtype: list
        """
        return []


def has_aliases(o) -> bool:
    """
    Checks whether the object has any aliases.

    :param o: the object to check
    :return: True if at least one alias defined
    """
    return len(get_aliases(o)) > 0


def get_aliases(o) -> List[str]:
    """
    Returns the aliases of the object, if it implements AliasSupporter.

    :param o: the object to get the aliases for
    :return: the list of aliases
    :rtype: list
    """
    result = []
    if isinstance(o, AliasSupporter):
        aliases = o.aliases()
        if aliases is not None:
            result.extend(aliases)
    return result


def get_all_names(o: Plugin) -> List[str]:
    """
    Returns all the names for the plugin, including any aliases.

    :param o: the plugin to get the names for
    :type o: Plugin
    :return: the list of aliases
    :rtype: list
    """
    result = [o.name()]
    result.extend(get_aliases(o))
    return result


class LoggingHandler(object):
    """
    Mixin for classes that support logging
    """

    def logger(self) -> logging.Logger:
        """
        Returns the logger instance to use.

        :return: the logger
        :rtype: logging.Logger
        """
        raise NotImplementedError()


class PluginWithLogging(Plugin, LoggingHandler):
    """
    Plugin that handles logging.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the handler.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__()
        self.logging_level = logging_level
        self.logger_name = logger_name
        self._logger = None

    def logger(self) -> logging.Logger:
        """
        Returns the logger instance to use.

        :return: the logger
        :rtype: logging.Logger
        """
        if self._logger is None:
            if (self.logger_name is not None) and (len(self.logger_name) > 0):
                logger_name = self.logger_name
            else:
                logger_name = self.name()
            self._logger = logging.getLogger(logger_name)
            set_logging_level(self._logger, self.logging_level)
        return self._logger

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        add_logging_level(parser)
        add_logger_name(parser, help_str="The custom name to use for the logger, uses the plugin name by default")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.logging_level = ns.logging_level
        self.logger_name = ns.logger_name
        self._logger = None


class OutputProducer(object):
    """
    Mixin for classes that generate output.
    """

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        raise NotImplementedError()


class InputConsumer(object):
    """
    Mixin for classes that consume input.
    """

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        raise NotImplementedError()


class Initializable(LoggingHandler):
    """
    Mixin for classes that require initialization and finalization
    """

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        self.logger().info("Initializing...")

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        self.logger().info("Finalizing...")


def init_initializable(handler: Initializable, handler_type: str, raise_again: bool = False) -> bool:
    """
    Initializes the commandline handler and outputs the stacktrace and help screen
    if it fails to do so. Optionally, the exception can be raised again (to propagate).

    :param handler: the handler to initialize
    :type handler: CommandlineHandler
    :param handler_type: the name of the type to use in the error message (eg "reader")
    :type handler_type: str
    :param raise_again: whether to raise the Exception again
    :type raise_again: bool
    :return: whether the initialization was successful
    :rtype: str
    """
    try:
        handler.initialize()
        return True
    except Exception as e:
        if isinstance(handler, Plugin):
            print("\nFailed to initialize %s '%s':\n" % (handler_type, handler.name()), file=sys.stderr)
        else:
            print("\nFailed to initialize %s '%s':\n" % (handler_type, get_class_name(handler)), file=sys.stderr)
        traceback.print_exc()
        if isinstance(handler, Plugin):
            print()
            handler.print_help()
            print()
        if raise_again:
            raise e
        return False


def check_compatibility(plugins: List[Plugin], match_all=AnyData):
    """
    Checks whether the plugins are compatible based on their inputs/outputs.
    Raises an exception if not compatible.

    :param plugins: the list of plugins to check
    :type plugins: list
    :param match_all: the class that always matches
    """
    if len(plugins) == 0:
        return

    for i in range(len(plugins) - 1):
        plugin1 = plugins[i]
        plugin2 = plugins[i + 1]
        index1 = "#" + str(i)
        index2 = "#" + str(i+1)
        if not isinstance(plugin1, OutputProducer):
            raise Exception(index1 + "/" + plugin1.name() + " is not an OutputProducer!")
        if not isinstance(plugin2, InputConsumer):
            raise Exception(index2 + "/" + plugin2.name() + " is not an InputConsumer!")
        classes1 = plugin1.generates()
        classes2 = plugin2.accepts()
        #
        if (match_all in classes1) or (match_all in classes2):
            return
        compatible = False
        for class1 in classes1:
            if class1 in classes2:
                compatible = True
                break
            for class2 in classes2:
                if issubclass(class1, class2):
                    compatible = True
                    break
            if compatible:
                break
        if not compatible:
            raise Exception(
                "Output(s) of " + index1 + "/" + plugin1.name()
                + " not compatible with input(s) of " + index2 + "/" + plugin2.name() + ": "
                + classes_to_str(classes1) + " != " + classes_to_str(classes2))



import math

from typing import Dict, List


def gcd(values: List[int]):
    """
    Computes the greatest common denominator for a list of integers.

    :param values: the integers to find the gcd for (at least 2)
    :type values: list
    :return: the gcd
    :rtype: int
    """
    if len(values) < 2:
        raise Exception("At least two numbers required for gcd!")
    values = values[:]
    result = values.pop(0)
    while len(values) > 0:
        b = values.pop(0)
        result = math.gcd(result, b)
    return result


class Splitter:
    """
    Class for dividing a token stream into sub-streams.
    """

    def __init__(self, split_ratios: List[int], split_names: List[str]):
        """
        Initializes the splitter.

        :param split_ratios: the list of (int) ratios to use (must sum up to 100)
        :type split_ratios: list
        :param split_names: the list of names to use for the splits
        :type split_names: list
        """
        self.split_ratios = split_ratios
        self.split_names = split_names
        self._schedule = None
        self._counter = 0
        self._stats = None

    def initialize(self):
        """
        Initializes the splitter, throws exceptions if undefined state or incorrect values.
        """
        if self.split_ratios is None:
            raise Exception("No split ratios defined!")
        if self.split_names is None:
            raise Exception("No split name defined!")
        if len(self.split_ratios) != len(self.split_names):
            raise Exception("Differing number of split ratios and names: %d != %d" % (len(self.split_ratios), len(self.split_names)))
        if sum(self.split_ratios) != 100:
            raise Exception("Split ratios must sum up to 100, but got: %d" % sum(self.split_ratios))

        # compute greatest common denominator and generate schedule
        _gcd = gcd(self.split_ratios)
        self._schedule = [0] * (len(self.split_ratios) + 1)
        for i, ratio in enumerate(self.split_ratios):
            self._schedule[i + 1] = self._schedule[i] + ratio / _gcd
        self._counter = 0
        self._stats = dict()

    def reset(self):
        """
        Resets the counter and stats.
        """
        self._counter = 0
        self._stats = dict()

    def next(self) -> str:
        """
        Returns the next split name.

        :return: the name of the split
        :rtype: str
        """
        split = None
        for i in range(len(self.split_names)):
            if (self._counter >= self._schedule[i]) and (self._counter < self._schedule[i + 1]):
                split = self.split_names[i]

        # update counter
        self._counter += 1
        if self._counter == self._schedule[-1]:
            self._counter = 0

        # update stats
        if split not in self._stats:
            self._stats[split] = 0
        self._stats[split] += 1

        return split

    def stats(self) -> Dict:
        """
        Returns the statistics.
        """
        return self._stats

    def counter(self) -> int:
        """
        Returns the counter.
        """
        return self._counter



import importlib
import inspect

from typing import Optional, Tuple, List


class AnyData:
    """
    Dummy class that can be used for matching anything in compatibility tests.
    """
    pass


def get_class(full_class_name: Optional[str] = None,
              module_name: Optional[str] = None, class_name: Optional[str] = None) -> type:
    """
    Returns the class object associated with the dot-notation classname.
    Either the fully qualified class name or module and class name can be supplied.

    :param full_class_name: the classname (module + name)
    :type full_class_name: str
    :param module_name: the name of the module
    :type module_name: str
    :param class_name: the
    :return: the class object
    """
    if full_class_name is not None:
        parts = full_class_name.split('.')
        module = ".".join(parts[:-1])
        m = importlib.import_module(module)
        c = getattr(m, parts[-1])
        return c
    elif (module_name is not None) and (class_name is not None):
        m = importlib.import_module(module_name)
        c = getattr(m, class_name)
        return c
    else:
        raise Exception("Either fully qualified class name or module and class name must be supplied!")


def fix_module_name(module: str, cls: str) -> Tuple[str, str]:
    """
    Turns a.b._C.C into a.b.C if possible.

    :param module: the module
    :type module: str
    :param cls: the class name
    :type cls: str
    :return: the (potentially) updated tuple of module and class name
    :rtype: tuple
    """
    if module.split(".")[-1].startswith("_"):
        try:
            module_short = ".".join(module.split(".")[:-1])
            getattr(importlib.import_module(module_short), cls)
            module = module_short
        except Exception:
            pass
    return module, cls


def get_class_name(obj: object) -> str:
    """
    Returns the fully qualified classname of the Python class or object.

    :param obj: the Python class/object to get the classname for
    :type obj: object
    :return: the generated classname
    :rtype: str
    """
    if inspect.isclass(obj):
        m, c = fix_module_name(obj.__module__, obj.__name__)
        return m + "." + c
    else:
        return get_class_name(obj.__class__)


def classes_to_str(classes: List):
    """
    Turns a list of classes into a string.

    :param classes: the list of classes to convert
    :type classes: list
    :return: the generated string
    :rtype: str
    """
    classes_str = list()
    for cls in classes:
        classes_str.append(get_class_name(cls))
    return ", ".join(classes_str)
